from enum import IntEnum
from typing import List, Tuple, Callable, Optional

import py_cui.keys

# primary keys
_PK_UP = py_cui.keys.KEY_UP_ARROW
_PK_RIGHT = py_cui.keys.KEY_RIGHT_ARROW
_PK_DOWN = py_cui.keys.KEY_DOWN_ARROW
_PK_LEFT = py_cui.keys.KEY_LEFT_ARROW
_PK_CONFIRM = py_cui.keys.KEY_SPACE
_PK_CANCEL = py_cui.keys.KEY_BACKSPACE
_PK_SITUATIONAL1 = py_cui.keys.KEY_CTRL_LEFT
_PK_SITUATIONAL2 = py_cui.keys.KEY_CTRL_RIGHT
_PK_PAUSE = py_cui.keys.KEY_TAB
_PK_HELP = py_cui.keys.KEY_H_LOWER
_PK_MATRIX_POPUP = py_cui.keys.KEY_M_LOWER

# secondary keys (alternatives to primary keys)
_SK_UP = [py_cui.keys.KEY_W_LOWER]
_SK_RIGHT = [py_cui.keys.KEY_D_LOWER]
_SK_DOWN = [py_cui.keys.KEY_S_LOWER]
_SK_LEFT = [py_cui.keys.KEY_A_LOWER]
_SK_CONFIRM = [py_cui.keys.KEY_ENTER]
_SK_CANCEL = [py_cui.keys.KEY_SHIFT_LEFT, py_cui.keys.KEY_A_UPPER]
_SK_SITUATIONAL1 = [py_cui.keys.KEY_Q_LOWER]
_SK_SITUATIONAL2 = [py_cui.keys.KEY_E_LOWER]
_SK_PAUSE = [py_cui.keys.KEY_P_LOWER]
_SK_HELP = []  # no alternative

# hot keys (not needed for playing but can be helpful/increase convenience)
_HOT_KEYS = [
    [py_cui.keys.KEY_1],
    [py_cui.keys.KEY_2],
    [py_cui.keys.KEY_3],
    [py_cui.keys.KEY_4],
    [py_cui.keys.KEY_5],
    [py_cui.keys.KEY_6],
    [py_cui.keys.KEY_7],
    [py_cui.keys.KEY_8],
    [py_cui.keys.KEY_9],
    [py_cui.keys.KEY_0, 94],  # 94 = ^
]

# debug keys (not used in normal gameplay)
_DK_RERENDER = [py_cui.keys.KEY_CTRL_R]
_DK_PRINT_SCREEN = [py_cui.keys.KEY_CTRL_P]
_DK_CHEAT_INPUT = [py_cui.keys.KEY_CTRL_I]
_DK_CHEAT_LIST = [py_cui.keys.KEY_CTRL_L]
_DK_STOP_SIM = [py_cui.keys.KEY_ESCAPE]


class Keys(IntEnum):
    # 0 - 3
    MoveUp = 0
    MoveRight = MoveUp + 1
    MoveDown = MoveUp + 2
    MoveLeft = MoveUp + 3

    # 4 - 7
    SelectionUp = MoveLeft + 1
    SelectionRight = SelectionUp + 1
    SelectionDown = SelectionUp + 2
    SelectionLeft = SelectionUp + 3

    # 8 - 13
    PopupClose = SelectionLeft + 1
    PopupUp = PopupClose + 1
    PopupDown = PopupClose + 2
    PopupRight = PopupClose + 3
    PopupLeft = PopupClose + 4
    PopupReopen = PopupClose + 5

    # 14 - 19
    Action = PopupReopen + 1
    Cancel = Action + 1
    Situational1 = Cancel + 1
    Situational2 = Situational1 + 1
    Help = Situational2 + 1

    Pause = Help + 1
    MatrixPopup = Pause + 1

    # 21 - 30
    HotKey1 = MatrixPopup + 1
    HotKey2 = HotKey1 + 1
    HotKey3 = HotKey1 + 2
    HotKey4 = HotKey1 + 3
    HotKey5 = HotKey1 + 4
    HotKey6 = HotKey1 + 5
    HotKey7 = HotKey1 + 6
    HotKey8 = HotKey1 + 7
    HotKey9 = HotKey1 + 8
    HotKey0 = HotKey1 + 9

    # 31 - 33
    Render = HotKey0 + 1
    PrintScreen = Render + 1
    StopSimulator = PrintScreen + 1

    # 34 - 35
    CheatInput = StopSimulator + 1
    CheatList = CheatInput + 1

    # non valid keys
    Invalid = 100
    ErrorMarker = 101
    LevelBegin = 110

    @staticmethod
    def invalid_values() -> "List[Keys]":
        return [Keys.Invalid, Keys.ErrorMarker, Keys.LevelBegin]

    @staticmethod
    def selection_keys() -> "List[Keys]":
        return [Keys.SelectionUp, Keys.SelectionRight, Keys.SelectionDown, Keys.SelectionLeft]

    @staticmethod
    def main_keys() -> "List[Keys]":
        return Keys.selection_keys() + [Keys.Action, Keys.Cancel]

    @staticmethod
    def hotkey(index: int) -> "Keys":
        hotkeys = Keys.hotkeys()
        if 0 <= index < len(hotkeys):
            return hotkeys[index]
        return Keys.Invalid

    @staticmethod
    def hotkeys() -> "List[Keys]":
        return [Keys.HotKey0, Keys.HotKey1, Keys.HotKey2, Keys.HotKey3, Keys.HotKey4, Keys.HotKey5, Keys.HotKey6,
                Keys.HotKey7, Keys.HotKey8, Keys.HotKey9]

    @staticmethod
    def from_code(code: int) -> "Keys":
        num = code - 1  # since code = num + 1
        return Keys.from_index(num)

    @staticmethod
    def from_index(index: int) -> "Keys":
        if index < len(Keys):
            for i, elem in enumerate(Keys):
                if i == index:
                    if elem != Keys(index):
                        from qrogue.util import Config
                        Config.check_reachability("Controls.from_index()-inequality")
                    return elem
            # performance of above version might be better?
            # values = [elem for elem in Keys]
            # return values[index]
        else:
            values = Keys.invalid_values()
            if index < len(values):
                return values[index]
        return Keys.Invalid

    def __init__(self, num):
        self.__num = num

    @property
    def num(self) -> int:
        return self.__num

    @property
    def code(self) -> int:
        # must be a valid char
        return self.__num + 1

    def to_char(self) -> str:
        return chr(self.code)


class Controls:
    INVALID_KEY = py_cui.keys.KEY_ALT_I  # is not allowed to be used as a valid key in the controls!
    __KEYS = [
        # move: 0 - 3
        [_PK_UP] + _SK_UP,
        [_PK_RIGHT] + _SK_RIGHT,
        [_PK_DOWN] + _SK_DOWN,
        [_PK_LEFT] + _SK_LEFT,
        # select: 4 - 7
        [_PK_UP] + _SK_UP,
        [_PK_RIGHT] + _SK_RIGHT,
        [_PK_DOWN] + _SK_DOWN,
        [_PK_LEFT] + _SK_LEFT,
        # popups: 8 - 13
        [_PK_CONFIRM] + _SK_CONFIRM + [_PK_CANCEL] + _SK_CANCEL,  # close
        [_PK_UP] + _SK_UP,
        [_PK_DOWN] + _SK_DOWN,
        [_PK_RIGHT] + _SK_RIGHT,
        [_PK_LEFT] + _SK_LEFT,
        [_PK_HELP],  # reopen
        # 14 - 19
        [_PK_CONFIRM] + _SK_CONFIRM,  # action
        [_PK_CANCEL] + _SK_CANCEL,  # cancel/back
        [_PK_SITUATIONAL1] + _SK_SITUATIONAL1,
        [_PK_SITUATIONAL2] + _SK_SITUATIONAL2,
        [_PK_HELP],  # actual help
        [_PK_PAUSE] + _SK_PAUSE,  # (Escape doesn't work here due to its special purpose for the engine)
        [_PK_MATRIX_POPUP],     # special key to open a matrix in a popup (e.g., if window is too narrow to display 3-qubit matrices in their corresponding widget)

        # special purpose hotkeys: 20 - 29
        _HOT_KEYS[0],
        _HOT_KEYS[1],
        _HOT_KEYS[2],
        _HOT_KEYS[3],
        _HOT_KEYS[4],
        _HOT_KEYS[5],
        _HOT_KEYS[6],
        _HOT_KEYS[7],
        _HOT_KEYS[8],
        _HOT_KEYS[9],

        # debugging keys: 30 - 32
        _DK_RERENDER,  # render screen
        _DK_PRINT_SCREEN,  # print screen
        _DK_STOP_SIM,  # stop simulator
        # cheat keys: 33 - 34
        _DK_CHEAT_INPUT,  # cheat input
        _DK_CHEAT_LIST,  # cheat list
    ]

    def __init__(self, handle_key_presses: Callable[[int], None]):
        self.__handle_key_presses = handle_key_presses
        self.__pycui_keys = Controls.__KEYS.copy()

    def encode(self, key_pressed: int) -> Keys:
        """
        Converts a pressed (raw) key to an internal (logical) Key-representation that encodes the corresponding action.
        :param key_pressed: the key that was pressed
        :return: an element of Key corresponding to the action executed by pressing key_pressed
        """
        for i in range(len(self.__pycui_keys)):
            if key_pressed in self.__pycui_keys[i]:
                return Keys.from_index(i)
        return Keys.Invalid

    def decode(self, key_code: int) -> Tuple[int, Keys]:
        """
        Decodes a code representation to a corresponding Keys element
        :param key_code: code representation of a Keys element
        :return: a keyboard key that can be pressed to trigger the corresponding action and its logical counterpart
        """
        if key_code == Keys.Invalid.code:
            return Controls.INVALID_KEY, Keys.Invalid
        key = Keys.from_code(key_code)
        return self.get_key(key), key

    def get_keys(self, key: Keys) -> List[int]:
        if key in Keys.invalid_values():
            return []
        return self.__pycui_keys[key.num]

    def get_key(self, key: Keys, index: int = 0) -> int:
        if key in Keys.invalid_values():
            return Controls.INVALID_KEY
        keys = self.get_keys(key)
        if 0 <= index < len(keys):
            return keys[index]
        return keys[0]

    def get_hotkey(self, number: int) -> List[int]:
        if number == 0:
            number = 10  # 0 is after 9, so at the 10th position
        base = Keys.HotKey1.num - 1
        return self.__pycui_keys[base + number]

    def handle(self, key: Keys) -> int:
        """
        :returns: the PyCUI-internal representation of the handled key
        """
        key_pressed = self.get_key(key)
        self.__handle_key_presses(key_pressed)
        return key_pressed

    @property
    def action(self) -> List[int]:
        return self.__pycui_keys[Keys.Action.num]

    def are_equivalent(self, key1: Keys, key2: Keys, raw_key: int) -> bool:
        # check if the pressed raw key is part of both logical keys, making them equivalent for the given press
        return raw_key in self.get_keys(key1) and raw_key in self.get_keys(key2)

    def to_keyboard_string(self, keys: Keys, index: Optional[int] = None, separator: str = ", ",
                           end_separator: Optional[str] = None) -> str:
        # does not work nicely for non-printable keyboard keys (e.g., Spacebar = ' ', Backspace = '\x08')
        # don't use keys.to_char() since it's only for internal use and doesn't correspond to the pressed keyboard key

        if index is None:  # use all possible keyboard keys
            chars = [chr(key) for key in self.get_keys(keys)]
            if len(chars) == 0: return "INVALID"
            if len(chars) == 1: return chars[0].upper()

            if end_separator is None: end_separator = separator
            # separate all chars and add the last one manually with its special end_separator
            return separator.join(chars[:-1]) + end_separator + chars[-1]

        else:
            key = self.get_key(keys, index)
            if key == Controls.INVALID_KEY: return "INVALID"
            return chr(key).upper()



import json
import math
from typing import List, Optional, Dict, Tuple, Iterable, Any, Callable

from qrogue.game.world.dungeon_generator import QrogueLevelGenerator
from qrogue.game.world.map import LessonMap, CallbackPack
from qrogue.game.world.map.rooms import AreaType
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.game.world.tiles import TileCode, Tile
from qrogue.util import RandomManager, MapConfig, MyRandom, Logger
from qrogue.util.util_functions import my_str
from .learnables import LearnableMap, LearnableRoom
from .wave_function import WaveFunction
from .wfc_learner import WFCLearner, WFCLearnMatrix


class WFCGenerator:
    @staticmethod
    def get_pos_weight_json_string(wfc_gen: "WFCGenerator") -> str:
        json_dicts = []
        for pos in wfc_gen.__learner.positions:
            json_dict = {}
            for key, value in wfc_gen.__learner.pos_weights(pos).items():
                json_dict[my_str(key)] = my_str(value)
            json_str = json.dumps(json_dict)
            json_dicts.append(f"\t\"{my_str(pos)}\": {json_str}")
        return "{\n" + ", \n".join(json_dicts) + "\n}"

    @staticmethod
    def get_type_weight_json_string(wfc_gen: "WFCGenerator") -> str:
        json_dicts = []
        for type_ in wfc_gen.__learner.types:
            json_dict = {}
            for key, value in wfc_gen.__learner.type_weights(type_).items():
                json_dict[my_str(key)] = my_str(value)
            json_str = json.dumps(json_dict)
            json_dicts.append(f"\t\"{my_str(type_)}\": {json_str}")
        return "{\n" + ", \n".join(json_dicts) + "\n}"

    @staticmethod
    def min_entropy(entropies: Dict[Coordinate, float]) -> Coordinate:
        min_pos = None
        min_val = -1
        for pos in entropies:
            val = entropies[pos]
            if min_val == -1 or val < min_val:
                min_pos = pos
                min_val = val
        return min_pos

    def __init__(self, pos_weights: Optional[Dict[Coordinate, Dict[Any, int]]] = None,
                 type_weights: Optional[Dict[Optional[Any], Dict[Any, int]]] = None,
                 data: Optional[List[WFCLearnMatrix]] = None):
        """
        Learns a wave function from the given data.
        """
        if pos_weights is None: pos_weights = {}
        if type_weights is None: type_weights = {}
        self.__learner = WFCLearner(pos_weights, type_weights)

        if data is not None:
            self.add_knowledge(data)

    def add_knowledge(self, data: List[WFCLearnMatrix]):
        assert len(data) > 0, "Cannot learn from empty list!"

        for matrix in data:
            self.__learner.learn(matrix)    # this updates pos_weights and type_weights
        self.__learner.remove_unwanted_values()

    def __weight(self, main: Any, neighbor: Any) -> float:
        """

        :param main: type of the main room
        :param neighbor: type of potential neighbor
        :return: normalized weight for the corresponding types
        """
        if main not in self.__learner.types or neighbor not in self.__learner.type_weights(main):
            return 0
        return self.__learner.type_weights(main)[neighbor] / sum(self.__learner.type_weights(main).values())

    def __weight_values(self, main: Any) -> Iterable[int]:
        if main not in self.__learner.types:
            return []
        return self.__learner.type_weights(main).values()

    def __weight_sum(self, main: Any) -> int:
        if main not in self.__learner.types:
            return 0
        return sum(self.__learner.type_weights(main).values())

    def __get_entropy(self, pos: Coordinate) -> float:
        if pos in self.__learner.positions:
            # shannon_entropy_for_square = log(sum(weight)) - (sum(weight * log(weight)) / sum(weight))
            weight_sum = sum(self.__learner.pos_weights(pos).values())
            log_sum = 0
            for w in self.__learner.pos_weights(pos).values():
                if w > 0:
                    log_sum += w * math.log(w)
            entropy = math.log(weight_sum) - (log_sum / weight_sum)
            return entropy
        return 0  # todo check which value to pick

    def generate(self, seed: int, width: Optional[int] = None, height: Optional[int] = None,
                 static_entries: Optional[Dict[Coordinate, Any]] = None,
                 is_assignable: Optional[Callable[[Any, Any], bool]] = None) -> List[List[Any]]:
        assert self.__learner.width > 0 and self.__learner.height > 0, \
            "Cannot generate without learning from data before!"

        rand: MyRandom = RandomManager.create_new(seed)

        # if no dimensions are given we take the learned one
        if width is None: width = self.__learner.width
        if height is None: height = self.__learner.height

        # create wave functions
        wave_functions: Dict[Coordinate, WaveFunction] = {}
        entropies: Dict[Coordinate, float] = {}
        for x in range(self.__learner.width):
            for y in range(self.__learner.height):
                c = Coordinate(x, y)
                entropies[c] = self.__get_entropy(c)
                if c in self.__learner.positions:
                    wave_functions[c] = WaveFunction(self.__learner.pos_weights(c))
                else:
                    wave_functions[c] = WaveFunction({})

        def propagate_collapse(position: Coordinate, collapsed_state: Any):
            # propagate collapse to all neighbors
            for direction in Direction.values():
                neighbor_pos = position + direction
                if 0 <= neighbor_pos.x < width and 0 <= neighbor_pos.y < height:
                    wave_functions[neighbor_pos].adapt_weights(self.__learner.type_weights(collapsed_state))

        if static_entries is not None:
            for pos in static_entries:
                value = static_entries[pos]
                if wave_functions[pos].force_value(value, is_assignable):
                    propagate_collapse(pos, wave_functions[pos].state)
                else:
                    Logger.instance().error(f"Failed to force value={value} for static entry at {pos}", show=False,
                                            from_pycui=False)

        while len(entropies) > 0:
            pos = WFCGenerator.min_entropy(entropies)
            value = wave_functions[pos].collapse(rand)
            propagate_collapse(pos, value)
            entropies.pop(pos)  # the position no longer needs to be handled

        matrix = [[None] * width for _ in range(height)]
        for pos in wave_functions:
            matrix[pos.y][pos.x] = wave_functions[pos].state
        return matrix

    def to_string(self, values_of_interest: Iterable[Any]) -> str:
        text = "WFC probabilities by feature type {\n"
        for main in values_of_interest:
            text += f"\t{my_str(main)} ("
            neighbor_texts = []
            for neighbor in values_of_interest:
                probability = self.__weight(main, neighbor)
                if probability > 0:
                    neighbor_texts.append(f"{my_str(neighbor)}: {probability * 100:2f}%")
            text += ", ".join(neighbor_texts)
            text += ")\n"
        text += "}"
        return text

    def __str__(self):
        return self.to_string(self.__learner.types)


def load_level(file_name: str, in_dungeon_folder: bool) -> Optional[LessonMap]:
    static_seed = 7  # seed does not matter since we are only interested in learning placements
    generator = QrogueLevelGenerator(lambda s: True, lambda s: None, lambda s, c: None, lambda s0, s1: None,
                                     CallbackPack.dummy())
    level, _ = generator.generate(static_seed, file_name, in_dungeon_folder)
    return level


class WFCLayoutGenerator(WFCGenerator):
    def __init__(self, templates: List[Tuple[str, bool]]):
        """
        Learns the wave function from the given templates and builds a level layout based on that.

        :param templates: list of tuples that define a filename and whether it can be found in the dungeon folder or not
        """
        self.__connection_weights: Dict[Optional[AreaType], Dict[Optional[AreaType], int]] = {}

        data: List[LearnableMap] = []
        for template in templates:
            # load levels
            filename, in_dungeon_folder = template
            level = load_level(filename, in_dungeon_folder)
            assert level is not None, f"Could not learn from file \"{filename}\""

            data.append(LearnableMap(level))

        super(WFCLayoutGenerator, self).__init__(data=data)

    def generate(self, seed: int, width: Optional[int] = None, height: Optional[int] = None,
                 static_entries: Optional[Dict[Coordinate, AreaType]] = None,
                 is_assignable: Optional[Callable[[AreaType, AreaType], bool]] = None) -> List[List[AreaType]]:
        # we need to first place a SpawnRoom as static entry
        if static_entries is None:
            static_entries = {}
        rm = RandomManager.create_new(seed)
        # todo random based on type_weights
        spawn_x = rm.get_int(0, MapConfig.map_width(), "spawn_x in WFCLayoutGenerator")
        spawn_y = rm.get_int(0, MapConfig.map_height(), "spawn_y in WFCLayoutGenerator")
        static_entries[Coordinate(spawn_x, spawn_y)] = AreaType.SpawnRoom

        return super(WFCLayoutGenerator, self).generate(seed, width, height, static_entries, is_assignable)


class WFCRoomGenerator(WFCGenerator):
    @staticmethod
    def get_level_list() -> List[Tuple[str, bool]]:
        return [(level, True) for level in MapConfig.level_list()[2:]]

    @staticmethod
    def from_level_files(templates: List[Tuple[str, bool]], room_type: AreaType) -> "WFCRoomGenerator":
        data: List[LearnableRoom] = []
        for template in templates:
            # load levels
            filename, in_dungeon_folder = template
            level = load_level(filename, in_dungeon_folder)
            assert level is not None, f"Could not learn from file \"{filename}\""

            # load rooms from the provided levels
            for x in range(level.width):
                for y in range(level.height):
                    room = level.room_at(x, y)
                    if room is not None and room.type == room_type:
                        data.append(LearnableRoom(room))
        return WFCRoomGenerator(room_type, data=data)

    @staticmethod
    def from_json_dicts(room_type: AreaType, json_pos: Dict[str, Dict[str, str]],
                        json_type: Dict[str, Dict[str, str]]) -> Optional["WFCRoomGenerator"]:
        pos_weights: Dict[Coordinate, Dict[LearnableRoom.TileData, int]] = {}
        for str_pos in json_pos:
            pos = Coordinate.from_string(str_pos)
            if pos is None: return None     # todo: maybe just skip instead? the remaining data might still be valuable

            pos_weights[pos] = {}
            for key, value in json_pos[str_pos].items():
                tile_data = LearnableRoom.TileData.from_string(key)
                weight = int(value)
                pos_weights[pos][tile_data] = weight

        type_weights: Dict[Optional[LearnableRoom.TileData], Dict[LearnableRoom.TileData, int]] = {}
        for str_type in json_type:
            tile_data = LearnableRoom.TileData.from_string(str_type)
            type_weights[tile_data] = {}
            for key, value in json_type[str_type].items():
                neighbor_data = LearnableRoom.TileData.from_string(key)
                weight = int(value)
                type_weights[tile_data][neighbor_data] = weight

        return WFCRoomGenerator(room_type, pos_weights, type_weights)

    def __init__(self, room_type: AreaType, pos_weights: Optional[Dict[Coordinate, Dict[Any, int]]] = None,
                 type_weights: Optional[Dict[Optional[Any], Dict[Any, int]]] = None,
                 data: Optional[List[WFCLearnMatrix]] = None):
        self.__room_type = room_type
        super(WFCRoomGenerator, self).__init__(pos_weights, type_weights, data)

    @property
    def room_type(self) -> AreaType:
        return self.__room_type

    def generate(self, seed: int, width: Optional[int] = None, height: Optional[int] = None,
                 static_entries: Optional[Dict[Coordinate, Tile]] = None,
                 is_assignable: Optional[Callable[[LearnableRoom.TileData, Tile], bool]] = None) \
            -> List[List[LearnableRoom.TileData]]:
        if is_assignable is None:
            def is_assignable(key: LearnableRoom.TileData, value: Tile) -> bool:
                return key.code == value.code
        return super().generate(seed, width, height, static_entries, is_assignable)


class WFCEmptyRoomGenerator(WFCGenerator):
    def generate(self, seed: Optional[int] = None, width: Optional[int] = None, height: Optional[int] = None,
                 static_entries: Optional[Dict[Coordinate, Any]] = None,
                 is_assignable: Optional[Callable[[LearnableRoom.TileData, Tile], bool]] = None) \
            -> List[List[LearnableRoom.TileData]]:
        if width is None: width = MapConfig.room_width()
        if height is None: height = MapConfig.room_height()
        # return an empty room (only Floor tiles)
        return [[LearnableRoom.TileData(TileCode.Floor, None) for _ in range(width)] for _ in range(height)]



import enum
import os
import pathlib
from typing import Tuple, Optional

from qrogue.util.util_functions import cur_datetime


class FileTypes(enum.Enum):
    Log = ".qrlog"
    KeyLog = ".qrkl"
    ScreenPrint = ".qrsc"
    Save = ".qrsave"
    Dungeon = ".qrdg"
    World = ".qrw"
    Templates = ".txt"


class PathConfig:
    __DATA_FOLDER = "data"
    __DEFAULT_GAME_DATA_PATH = os.path.join(os.path.dirname(__file__), "..", "..", __DATA_FOLDER)
    __DEFAULT_USER_DATA_PATH = os.path.join(os.path.dirname(__file__), "..", "..", "QrogueData")
    __LOG_FOLDER = "logs"
    __KEY_LOG_FOLDER = "keylogs"
    __SCREEN_PRINTS_FOLDER = "screenprints"
    __SAVE_DATA_FOLDER = "saves"
    __DUNGEON_FOLDER = "dungeons"
    __TEMPLATE_ROOMS = "rooms"
    __TEMPLATE_HALLWAYS = "hallways"
    __TEMPLATE_STV_POOLS = "stv_pools"
    __TEMPLATE_REWARD_POOLS = "reward_pools"
    __TEMPLATE_FILE = f"templates{FileTypes.Templates.value}"
    __SAVE_FILE_PREFIX = "qrogue-save"
    __FRESH_SAVE_FILE = "fresh"
    __SAVE_FILE_NUMERATION_SEPARATOR = "_"
    __NUMBER_OF_SAVE_FILES = 7  # how many save files can be present before we delete the oldest one

    __Base_Path = __DEFAULT_GAME_DATA_PATH
    __User_Data_Path = __DEFAULT_USER_DATA_PATH

    @staticmethod
    def __now_str() -> str:
        return cur_datetime().strftime("%d%m%Y_%H%M%S")

    @staticmethod
    def __save_file_str(num: int) -> str:
        return os.path.join(PathConfig.__SAVE_DATA_FOLDER,
                            f"{PathConfig.__SAVE_FILE_PREFIX}{PathConfig.__SAVE_FILE_NUMERATION_SEPARATOR}{num}"
                            f"{FileTypes.Save.value}")

    @staticmethod
    def __get_save_files_stats() -> Tuple[int, int]:
        """

        :return: number of available save files, number of the latest save file
        """
        files = os.listdir(PathConfig.user_data_path(PathConfig.__SAVE_DATA_FOLDER))
        num = -1
        num_of_files = 0
        for file in files:
            if file.startswith(PathConfig.__SAVE_FILE_PREFIX) and file.endswith(FileTypes.Save.value):
                file_ending_index = len(file) - len(FileTypes.Save.value)
                num_of_files += 1
                cur_num = int(file[file.rindex(PathConfig.__SAVE_FILE_NUMERATION_SEPARATOR) + 1:file_ending_index])
                if cur_num > num:
                    num = cur_num
        return num_of_files, num

    @staticmethod
    def launch_config_path() -> str:
        return os.path.join(PathConfig.__Base_Path, "qrogue_launch.config")

    @staticmethod
    def default_base_path() -> str:
        return PathConfig.__DEFAULT_GAME_DATA_PATH

    @staticmethod
    def default_user_data_path() -> str:
        return PathConfig.__DEFAULT_USER_DATA_PATH

    @staticmethod
    def keylog_folder() -> str:
        return PathConfig.__KEY_LOG_FOLDER

    @staticmethod
    def create_folder_structure(user_data_path: str) -> None:
        pathlib.Path(user_data_path).mkdir(parents=True, exist_ok=True)
        log_path = os.path.join(user_data_path, PathConfig.__LOG_FOLDER)
        key_log_path = os.path.join(user_data_path, PathConfig.__KEY_LOG_FOLDER)
        screen_prints_path = os.path.join(user_data_path, PathConfig.__SCREEN_PRINTS_FOLDER)
        save_data_path = os.path.join(user_data_path, PathConfig.__SAVE_DATA_FOLDER)

        if not os.path.exists(log_path):
            os.mkdir(log_path)
        if not os.path.exists(key_log_path):
            os.mkdir(key_log_path)
        if not os.path.exists(screen_prints_path):
            os.mkdir(screen_prints_path)
        if not os.path.exists(save_data_path):
            os.mkdir(save_data_path)

    @staticmethod
    def set_base_path(base_path: str):
        if os.path.exists(base_path):
            PathConfig.__Base_Path = base_path
        else:
            raise FileNotFoundError(f"Given base path is not valid: {base_path} [normalized value = "
                                    f"{os.path.normpath(base_path)}]")

    @staticmethod
    def set_user_data_path(user_data_path: str):
        PathConfig.create_folder_structure(user_data_path)  # does nothing for existing folders
        PathConfig.__User_Data_Path = user_data_path

    @staticmethod
    def load_paths(custom_data_path: Optional[str], custom_user_data_path: Optional[str]) -> bool:
        try:
            if custom_data_path is None or custom_user_data_path is None:
                if os.path.exists(PathConfig.launch_config_path()):
                    with open(PathConfig.launch_config_path()) as f:
                        content = f.readlines()
                else:
                    content = None

            if custom_data_path:
                data_path = custom_data_path
            elif content:
                data_path = content[1]
            else:
                # use the default if nothing is provided
                data_path = PathConfig.default_base_path()

            if custom_user_data_path:
                user_data_path = custom_user_data_path
            elif content:
                user_data_path = content[2]
            else:
                # use the default if nothing is provided
                user_data_path = PathConfig.default_user_data_path()

            if data_path == "\n":
                data_path = PathConfig.default_base_path()
            elif data_path.endswith("\n"):
                data_path = data_path[:-1]

            if user_data_path == "\n":
                user_data_path = PathConfig.default_user_data_path()
            elif user_data_path.endswith("\n"):
                user_data_path = user_data_path[:-1]

            PathConfig.set_base_path(data_path)
            PathConfig.set_user_data_path(user_data_path)
            return os.path.exists(PathConfig.__Base_Path) and os.path.exists(PathConfig.__User_Data_Path)

        except Exception as error:
            print(error)
            return False

    @staticmethod
    def base_path(file_name: str = "") -> str:
        return os.path.join(PathConfig.__Base_Path, file_name)

    @staticmethod
    def user_data_path(file_name: str = "") -> str:
        return os.path.join(PathConfig.__User_Data_Path, file_name)

    @staticmethod
    def new_log_file() -> str:
        path = os.path.join(PathConfig.__LOG_FOLDER, f"{PathConfig.__now_str()}")

        # this if should in general be True 99% of the time, and 100% of the time on regular usage
        if not os.path.exists(path + f"{FileTypes.Log.value}"):
            return path + f"{FileTypes.Log.value}"

        # for the very small chance that another log file has the same datetime (up to seconds), we add a simple
        # counter suffix
        counter = 0
        while os.path.exists(path + f"_{counter}{FileTypes.Log.value}"): counter += 1
        return path + f"_{counter}{FileTypes.Log.value}"

    @staticmethod
    def new_key_log_file(seed: int, is_level: bool = True) -> str:
        now_str = PathConfig.__now_str()
        if is_level:
            return os.path.join(PathConfig.__KEY_LOG_FOLDER, f"{now_str}_seed{seed}{FileTypes.KeyLog.value}")
        else:
            return os.path.join(PathConfig.__KEY_LOG_FOLDER, f"{now_str}_meta{seed}{FileTypes.KeyLog.value}")

    @staticmethod
    def get_seed_from_key_log_file(file_name: str) -> Optional[int]:
        path = PathConfig.user_data_path(os.path.join(PathConfig.__KEY_LOG_FOLDER, file_name))
        content = PathConfig.read(path, True)
        start = content.index("Seed=") + len("Seed=")
        end = content.index("\n", start)
        seed = content[start:end]
        return int(seed)

    @staticmethod
    def new_screen_print(text: str):
        now_str = PathConfig.__now_str()
        path = os.path.join(PathConfig.__SCREEN_PRINTS_FOLDER, f"{now_str}{FileTypes.ScreenPrint.value}")
        PathConfig.write(path, now_str + "\n" + text, may_exist=True, append=True)

    @staticmethod
    def new_save_file(text: str):
        num_of_files, num = PathConfig.__get_save_files_stats()

        # increment to get the highest number for the new save file (if no save file exists yet, -1 will be incremented
        # to 0)
        num += 1
        if num_of_files >= PathConfig.__NUMBER_OF_SAVE_FILES:
            oldest_num = num - PathConfig.__NUMBER_OF_SAVE_FILES
            PathConfig.delete(PathConfig.__save_file_str(oldest_num))

        file_path = PathConfig.__save_file_str(num)
        try:
            PathConfig.write(file_path, text, may_exist=False)
        except Exception as ex:
            PathConfig.delete(file_path)    # delete the newly created file because it is corrupt
            raise ex

    @staticmethod
    def write_auto_save(text: str):
        # don't use the same ending because it would mess with __get_save_files_stats()
        file_path = os.path.join(PathConfig.__SAVE_DATA_FOLDER,
                                 f"{PathConfig.__SAVE_FILE_PREFIX}{FileTypes.Save.value}_auto")
        PathConfig.write(file_path, text, may_exist=True)

    @staticmethod
    def find_latest_save_file() -> str:
        _, num = PathConfig.__get_save_files_stats()
        if num >= 0:
            return PathConfig.__save_file_str(num)
        else:
            return PathConfig.base_path(f"{PathConfig.__FRESH_SAVE_FILE}{FileTypes.Save.value}")

    @staticmethod
    def read_keylog_buffered(file_name: str, in_keylog_folder: bool = True, buffer_size: int = 1024) -> str:
        if not file_name.endswith(FileTypes.KeyLog.value):
            file_name += FileTypes.KeyLog.value
        if in_keylog_folder:
            path = PathConfig.user_data_path(os.path.join(PathConfig.__KEY_LOG_FOLDER, file_name))
        else:
            path = file_name
        if os.path.exists(path):
            with open(path, "r") as file:
                data = file.read(buffer_size)
                while data:
                    yield data
                    data = file.read(buffer_size)
        else:
            raise FileNotFoundError(f"There is no such key log file: {path}")

    @staticmethod
    def read_world(file_name: str, in_dungeon_folder: bool = True):
        if not file_name.endswith(FileTypes.World.value):
            file_name += FileTypes.World.value

        if in_dungeon_folder:
            path = PathConfig.base_path(os.path.join(PathConfig.__DUNGEON_FOLDER, file_name))
        else:
            path = file_name
        return PathConfig.read(path, in_user_path=False)

    @staticmethod
    def read_level(file_name: str, in_dungeon_folder: bool = True):
        if not file_name.endswith(FileTypes.Dungeon.value):
            file_name += FileTypes.Dungeon.value

        if in_dungeon_folder:
            path = PathConfig.base_path(os.path.join(PathConfig.__DUNGEON_FOLDER, file_name))
        else:
            path = file_name
        return PathConfig.read(path, in_user_path=False)

    @staticmethod
    def read(file_name: str, in_user_path: bool) -> str:
        if in_user_path:
            path = PathConfig.user_data_path(file_name)
        else:
            path = file_name
        if os.path.exists(path):
            with open(path, "r") as file:
                content = file.read()
            return content
        else:
            raise FileNotFoundError(f"File \"{file_name}\" could not be found!")

    @staticmethod
    def write(path: str, text: str, in_user_path: bool = True, may_exist: bool = True, append: bool = False):
        if in_user_path:
            path = PathConfig.user_data_path(path)  # data in base_path is static so we can only write user data
        mode = "x"
        if may_exist:
            if os.path.exists(path):
                mode = "w"
                if append:
                    mode = "a"
        with open(path, mode) as file:
            file.write(text)

    @staticmethod
    def delete(path: str):
        path = PathConfig.user_data_path(path)  # data in base_path is static so we can only delete user data
        if os.path.exists(path):
            os.remove(path)



from enum import IntEnum
from typing import Callable, Optional, List, Tuple

from qrogue.game.logic import Message
from qrogue.util import Config, PopupConfig, Logger, HelpText


class Popup:
    class Pos(IntEnum):
        Center = 0
        Top = 1
        TopRight = 2
        Right = 3
        BottomRight = 4
        Bottom = 5
        BottomLeft = 6
        Left = 7
        TopLeft = 8

        Matrix = 9

    class Dimension:
        def __init__(self, tiny: int, small: int, medium: int, large: int):
            self.__tiny = tiny
            self.__small = small
            self.__medium = medium
            self.__large = large

        @property
        def tiny(self) -> int:
            return self.__tiny

        @property
        def small(self) -> int:
            return self.__small

        @property
        def medium(self) -> int:
            return self.__medium

        @property
        def large(self) -> int:
            return self.__large

    # dimensions are without padding, so purely the size of the content part of a Popup
    DimWidth = Dimension(40, 70, 104, 140)
    DimHeight = Dimension(3, 5, 8, 13)

    __DEFAULT_POS = PopupConfig.default_position()
    __show_popup: Optional[Callable[[str, str, int, int, Optional[Tuple[int, int]], Optional[bool], Optional[int],
                                     Optional[PopupConfig.Importance]],
                                    None]] = None
    __check_achievement_callback: Optional[Callable[[str], bool]] = None
    __popup_queue: List["Popup"] = []
    __cur_popup: Optional["Popup"] = None

    @staticmethod
    def is_initialized() -> bool:
        return Popup.__show_popup is not None and Popup.__check_achievement_callback is not None

    @staticmethod
    def update_popup_functions(show_popup_callback: Callable[[str, str, int, int, Optional[Tuple[int, int]],
                                                              Optional[bool], Optional[int]], None]) -> None:
        """
        callable(title, text, color, position, dimensions, reopen, padding x)
        """
        Popup.__show_popup = show_popup_callback

    @staticmethod
    def update_check_achievement_function(check_achievement_callback: Callable[[str], bool]) -> None:
        Popup.__check_achievement_callback = check_achievement_callback

    @staticmethod
    def reset_queue():
        Popup.__popup_queue.clear()

    @staticmethod
    def on_close() -> bool:
        if Popup.__cur_popup is not None:
            Popup.__cur_popup.on_close_callback()
            Popup.__cur_popup.__on_close_callback = None  # clear callback to not execute it when reopening!
        Popup.__cur_popup = None
        if len(Popup.__popup_queue) > 0:
            next_popup = Popup.__popup_queue.pop(0)
            next_popup.show()
            return False  # don't fully close popup
        return True  # popup no longer needed so we can fully close it

    @staticmethod
    def message(title: str, text: str, reopen: bool, pos: Optional[int] = None,
                color: int = PopupConfig.default_color(), overwrite: bool = False,
                on_close_callback: Callable[[], None] = None,
                dimensions: Optional[Tuple[int, int]] = None,
                importance: Optional[PopupConfig.Importance] = None):
        if pos is None:
            pos = Popup.__DEFAULT_POS
        if importance is None:
            importance = PopupConfig.Importance.Undefined
        Popup(title, text, pos, color, reopen=reopen, show=True, overwrite=overwrite,
              on_close_callback=on_close_callback, dimensions=dimensions, importance=importance)

    @staticmethod
    def generic_info(title: str, text: str, reopen: Optional[bool] = None, pos: Optional[int] = None):
        if reopen is None:
            reopen = False
        if pos is None:
            pos = Popup.__DEFAULT_POS
        Popup.message(title, text, reopen=reopen, pos=pos, importance=PopupConfig.Importance.Info)

    @staticmethod
    def show_help(help_text: HelpText):
        Popup.generic_info(help_text.name, help_text.text)

    @staticmethod
    def error(text: str, reopen: Optional[bool] = None, pos: Optional[int] = None, overwrite: bool = True,
              log_error: bool = True, add_report_note: bool = False):
        """
        Args:
            text: a text describing the error
            reopen: whether the popup should be reopen-able (defaults to False)
            pos: position of the popup
            overwrite: whether this popup should overwrite the currently displayed one if there is any (defaults to True)
            log_error: whether we should forward the error to our Logger (defaults to True)
            add_report_note: whether to add a text to inform the player how to report the error
        """
        if reopen is None:
            reopen = False
        if pos is None:
            pos = Popup.__DEFAULT_POS
        if log_error: Logger.instance().error(text, show=False, from_pycui=False)
        if add_report_note:
            text += f"\nPlease report this error together with your user data to {Config.report_address()}"
        Popup.message("Error", text, reopen, pos, overwrite=overwrite, importance=PopupConfig.Importance.Error)

    @staticmethod
    def system_says(text: str, reopen: Optional[bool] = None, pos: Optional[int] = None):
        if reopen is None:
            reopen = True
        if pos is None:
            pos = Popup.__DEFAULT_POS
        Popup.message(Config.system_name(), text, reopen=reopen, pos=pos, importance=PopupConfig.Importance.Dialogue)

    @staticmethod
    def examiner_says(text: str, reopen: Optional[bool] = None, pos: Optional[int] = None):
        if reopen is None:
            reopen = True
        if pos is None:
            pos = Popup.__DEFAULT_POS
        Popup.message(Config.examiner_name(), text, reopen=reopen, pos=pos, importance=PopupConfig.Importance.Dialogue)

    @staticmethod
    def scientist_says(text: str, reopen: Optional[bool] = None, pos: Optional[int] = None,
                       dimensions: Optional[Tuple[int, int]] = None):
        if reopen is None:
            reopen = True
        if pos is None:
            pos = Popup.__DEFAULT_POS
        Popup.message(Config.scientist_name(), text, reopen=reopen, pos=pos, dimensions=dimensions,
                      importance=PopupConfig.Importance.Dialogue)

    @staticmethod
    def npc_says(name: str, text: str, reopen: Optional[bool] = None, pos: Optional[int] = None):
        if reopen is None:
            reopen = True
        if pos is None:
            pos = Popup.__DEFAULT_POS
        Popup.message(name, text, reopen=reopen, pos=pos, importance=PopupConfig.Importance.Dialogue)

    @staticmethod
    def from_message(message: Message, overwrite: bool = False):
        if Popup.__check_achievement_callback:
            ret = message.get(Popup.__check_achievement_callback)  # resolve possible alternative messages
            if ret is not None:
                title, text = ret
                reopen = message.priority
                Popup.message(title, text, reopen=reopen, pos=message.position, overwrite=overwrite,
                              importance=PopupConfig.Importance.Dialogue)

    @staticmethod
    def from_message_trigger(message: Message, on_close_callback: Callable[[], None]):
        if Popup.__check_achievement_callback:
            ret = message.get(Popup.__check_achievement_callback)  # resolve possible alternative messages
            if ret is not None:
                title, text = ret
                if Config.debugging():
                    title += f" {{@*{message.id}}}"
                reopen = message.priority
                Popup.message(title, text, reopen, pos=message.position, on_close_callback=on_close_callback,
                              importance=PopupConfig.Importance.Dialogue)

    @staticmethod
    def show_matrix(title: str, text: str, prefix: str = ""):
        # remove the additional whitespace at the end of a line (needed for CircuitMatrixWidget to center correctly,
        # but not here since a popup is left-aligned [although we are technically centered since we fit the width])
        lines = [line.rstrip(' ') for line in text.splitlines(keepends=False)]
        height = len(lines)
        width = max([len(line) for line in lines])
        Popup(title, prefix + "\n".join(lines), Popup.Pos.Matrix, PopupConfig.default_color(), reopen=False, show=True,
              overwrite=False, on_close_callback=None, dimensions=(height, width), padding_x=0,
              importance=PopupConfig.Importance.Info)

    def __init__(self, title: str, text: str, position: int, color: int = PopupConfig.default_color(),
                 show: bool = True, overwrite: bool = False, reopen: bool = True,
                 on_close_callback: Callable[[], None] = None, dimensions: Optional[Tuple[int, int]] = None,
                 padding_x: Optional[int] = None,
                 importance: PopupConfig.Importance = PopupConfig.Importance.Undefined):
        self.__title = title
        self.__text = text
        self.__position = position
        self.__dimensions = dimensions
        self.__padding_x = padding_x
        self.__importance = importance
        self.__color = color
        self.__reopen = reopen  # whether this popup should be reopen-able or not
        self.__on_close_callback = on_close_callback
        if show:
            self.show(overwrite)

    @property
    def _title(self) -> str:
        return self.__title

    @property
    def _text(self) -> str:
        return self.__text

    @property
    def _color(self) -> int:
        return self.__color

    @property
    def is_reopenable(self) -> bool:
        return self.__reopen

    def on_close_callback(self):
        if self.__on_close_callback:
            self.__on_close_callback()

    def _base_show(self):
        Popup.__show_popup(self.__title, self.__text, self.__position, self.__color, self.__dimensions, self.__reopen,
                           self.__padding_x, self.__importance)

    def _enqueue(self):
        Popup.__popup_queue.append(self)

    def show(self, overwrite: bool = False) -> None:
        if overwrite:
            Popup.__popup_queue.clear()
            Popup.__cur_popup = None
        if self.__cur_popup:
            self._enqueue()
        else:
            Popup.__cur_popup = self
            self._base_show()


class ConfirmationPopup(Popup):
    __DEFAULT_POS = PopupConfig.default_position()
    __show_popup: Callable[[str, str, int, Callable[[int], None], Optional[List[str]]], None] = None

    @staticmethod
    def update_popup_function(show_popup_callback: Callable[[str, str, int, Callable[[int], None], Optional[List[str]]],
                                                            None]):
        ConfirmationPopup.__show_popup = show_popup_callback

    @staticmethod
    def ask(title: str, text: str, callback: Callable[[int], None], answers: Optional[List[str]] = None):
        ConfirmationPopup(title, text, callback, answers)

    def __init__(self, title: str, text: str, callback: Callable[[int], None], answers: Optional[List[str]] = None,
                 position: int = __DEFAULT_POS, color: int = PopupConfig.default_color(), show: bool = True,
                 overwrite: bool = False):
        def on_close():
            callback(self.__confirmed)

        self.__answers = answers
        super().__init__(title, text, position, color, show, overwrite, reopen=False,
                         on_close_callback=on_close)
        self.__confirmed: Optional[int] = None

    def __set_confirmation(self, confirmed: int):
        # by setting confirmed here we determine the parameter of the callback called after closing
        self.__confirmed = confirmed

    def _base_show(self) -> None:
        ConfirmationPopup.__show_popup(self._title, self._text, self._color, self.__set_confirmation, self.__answers)



from typing import Optional


class ErrorConfig:
    @staticmethod
    def singleton(singleton: str) -> str:
        return f"The class \"{singleton}\" is a singleton!"

    @staticmethod
    def singleton_no_init(singleton: str) -> str:
        return f"The singleton \"{singleton}\" has not been initialized yet!"

    @staticmethod
    def singleton_reset(singleton: str) -> str:
        return f"Can only reset the singleton \"{singleton}\" during testing!"

    @staticmethod
    def invalid_map(map_name: str, intro_text: Optional[str] = None) -> str:
        if intro_text is None:
            intro_text = f"Could not find map \"{map_name}\"! "
        return intro_text + "Please download the game files again.\n" \
                            "If this error still occurs but you're sure that the corresponding file is present:"

    @staticmethod
    def qubit_overflow(actual_qubits: int, available_qubits: int) -> str:
        return f"Cannot place a gate with {actual_qubits} qubits on a circuit with only {available_qubits} qubit!\n" \
               f"Please press \"Cancel\" to abort placement."

    @staticmethod
    def raise_deletion_exception():
        """
        Raises a common exception to indicate that this is old code that needs to be removed and worked around.

        Returns: None

        """
        raise Exception("No longer usable!")



import os

from setuptools import setup

README_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'Readme.md')
with open(README_PATH) as readme_file:
    README = readme_file.read()
    # fix paths in Readme (PyPI has different base folder than GitHub)
    WRONG_PATH = "src=\"./images"
    CORRECT_PATH = "src=\"https://github.com/7Magic7Mike7/Qrogue/raw/master/images"
    start_index = 0
    while True:
        try:
            path_index = README.index(WRONG_PATH, start_index)
            start_index = path_index + len(WRONG_PATH)
            README = README[:path_index] + CORRECT_PATH + README[start_index:]
        except ValueError:
            break  # no more paths were found

setup(
    name='qrogue',  # Name of package folder
    packages=[
        'qrogue',
        'qrogue.util',
        'qrogue.util.config',
        'qrogue.game',
        'qrogue.game.logic',
        'qrogue.game.logic.base',
        'qrogue.game.logic.collectibles',
        'qrogue.game.logic.actors',
        'qrogue.game.logic.actors.controllables',
        'qrogue.game.logic.actors.puzzles',
        'qrogue.game.world',
        'qrogue.game.world.navigation',
        'qrogue.game.world.tiles',
        'qrogue.game.world.map',
        'qrogue.game.world.dungeon_generator',
        'qrogue.game.world.dungeon_generator.dungeon_parser',
        'qrogue.game.world.dungeon_generator.wave_function_collapse',
        'qrogue.graphics',
        'qrogue.graphics.rendering',
        'qrogue.graphics.popups',
        'qrogue.graphics.widgets',
        'qrogue.management',
        'qrogue.management.save_grammar',
    ],
    package_data={"data": ["qrogue/data"]},
    include_package_data=True,
    version='0.8.1.4',
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='QRogue is an educational Quantum Computing puzzle game inspired by the classical game Rogue.',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Michael Artner',
    author_email='michael.artner@hotmail.de',
    url='https://github.com/7Magic7Mike7/Qrogue',
    download_url='https://github.com/7Magic7Mike7/Qrogue/releases/tag/0.8.1',  # link to Release on GitHub
    keywords=['Quantum Computing', 'Gamification', 'Video Game', 'PyCUI'],
    install_requires=[
        'antlr4-python3-runtime==4.13.2',
        'numpy==1.22.3',
        'py-cui==0.1.4',
        'qiskit==0.34.2',
        'qiskit-aer==0.10.3',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Games/Entertainment :: Puzzle Games',
        'Programming Language :: Python :: 3.8',
    ],
    entry_points={
        'console_scripts': ['qrogue=qrogue.qrogue:start_qrogue'],
    }
)



from typing import Tuple, Optional

from qrogue.game.logic.base import StateVector, CircuitMatrix
from qrogue.game.logic.collectibles import Collectible
from qrogue.util import RandomManager, CheatConfig
from .target import Target


class Riddle(Target):
    """
    A Target with a restricted amount of edits to reach its StateVector.
    """

    __STABLE_PROBABILITY = 0.5

    def __init__(self, id_: int, target: StateVector, reward: Collectible, seed: int, edits: int = 1,
                 input_: Optional[StateVector] = None, stable_probability: float = __STABLE_PROBABILITY):
        """
        :param stable_probability: how likely it is to stay stable after all edits are used up
            (0=no edits possible, 1=infinitely many additional edits possible)
        """
        super().__init__(id_, target, reward, input_)
        self.__edits = edits
        self.__stable_probability = stable_probability
        self.__rm = RandomManager.create_new(seed)
        self.__can_attempt = True

    @property
    def edits(self) -> int:
        """

        :return: the number of edits that are left for reaching this Target's StateVector
        """
        return self.__edits

    @property
    def can_attempt(self) -> bool:
        if self.edits > 0 or CheatConfig.has_infinite_edits():
            self.__can_attempt = True
        else:
            self.__can_attempt = self.__rm.get() < self.__stable_probability
        return self.__can_attempt

    @property
    def is_active(self) -> bool:
        return super(Riddle, self).is_active and self.__can_attempt

    @property
    def flee_energy(self) -> int:
        """
        Fleeing (pausing) Riddles should always be possible, hence it costs 0 energy.

        :return: 0
        """
        return 0

    def is_reached(self, state_vector: StateVector, circ_matrix: CircuitMatrix) -> Tuple[bool, Optional[Collectible]]:
        success, reward = super(Riddle, self).is_reached(state_vector, circ_matrix)
        if success:
            assert reward is not None  # riddles always need to give a reward
        else:
            self.__edits = max(self.__edits - 1, 0)
            self.__can_attempt = self.can_attempt
        return success, reward



import qrogue.game
import qrogue.graphics
import qrogue.management
import qrogue.util
from qrogue.qrogue import setup_game
from qrogue.qrogue import simulate_game
from qrogue.qrogue import start_game
from qrogue.qrogue import start_qrogue
from qrogue.qrogue import validate_map



# exporting
from .QrogueLevelGenerator import QrogueLevelGenerator
from .generator import DungeonGenerator
from .random_generator import ExpeditionGenerator

# not exported:
# - .wave_function_collapse (should only be used by a generator)

# importing
# +graphics.popups
# +logic.actors
# +logic.actors.controllables
# +logic.actors.puzzles
# +logic.base
# +logic.collectibles
# +world.map
# +world.map.rooms
# +world.navigation
# +world.tiles
# +util



import enum
from typing import Callable, Tuple, List, Set, Optional

from qrogue.util.util_functions import open_folder
from .options import OptionsManager, Options
from .path_config import PathConfig
from .test_config import TestConfig
from .visual_config import PopupConfig, PyCuiColors


class MapConfig:
    @staticmethod
    def num_of_lessons() -> int:
        return 5  # Lesson 0 to 4

    @staticmethod
    def map_width() -> int:
        return 7

    @staticmethod
    def map_height() -> int:
        return 3

    @staticmethod
    def room_width() -> int:
        return 7

    @staticmethod
    def room_height() -> int:
        return MapConfig.room_width()

    @staticmethod
    def room_mid_x() -> int:
        return int(MapConfig.room_width() / 2)

    @staticmethod
    def room_mid_y() -> int:
        return int(MapConfig.room_height() / 2)

    @staticmethod
    def room_mid() -> Tuple[int, int]:
        return MapConfig.room_mid_x(), MapConfig.room_mid_y()

    @staticmethod
    def done_event_id() -> str:
        return "done"

    @staticmethod
    def specific_done_event_id(completed_map: str) -> str:
        return f"{completed_map}{MapConfig.done_event_id()}"

    @staticmethod
    def global_event_prefix() -> str:
        return "global_event_"

    @staticmethod
    def unlock_prefix() -> str:
        return "unlock_"

    @staticmethod
    def next_map_string() -> str:
        return "next"

    @staticmethod
    def back_map_string() -> str:
        return "back"

    @staticmethod
    def lesson_map_prefix() -> str:
        return "l"

    @staticmethod
    def expedition_map_prefix() -> str:
        return "expedition"

    @staticmethod
    def diff_code_separator() -> str:
        return "d"

    @staticmethod
    def puzzle_seed_separator() -> str:
        return "p"

    @staticmethod
    def tutorial_lesson_prefix() -> str:
        return "l0v"

    @staticmethod
    def first_uncleared() -> str:
        return "firstUncleared"

    @staticmethod
    def hub_world() -> str:
        return "worlds"

    @staticmethod
    def tutorial_world() -> str:
        return "w0"

    @staticmethod
    def first_world() -> str:
        return "w1"

    @staticmethod
    def intro_level() -> str:
        return f"{MapConfig.tutorial_lesson_prefix()}0"

    @staticmethod
    def exam() -> str:
        return "l0exam"

    @staticmethod
    def test_level() -> str:
        return "l0training"

    @staticmethod
    def level_list() -> List[str]:
        levels = []
        for knowledge_level in range(2):
            for i in range(MapConfig.num_of_lessons()):
                levels.append(f"l0k{knowledge_level}v{i}")
        # levels += ["l0training", "l0exam"]
        return levels


class CheatConfig:
    # __ALL and __NONE do not toggle! They simply set or reset all "real" cheats. All other/"real" cheats are toggles.
    __ALL = "aLL"
    __GOD_MODE = "Qod-Mode"     # puzzle checks always succeed, regardless of your circuit or additional constraints
    __INF_EDITS = "Supertry"    # riddles (e.g., boss puzzles) allow for infinite many edits
    __SCARED_RABBIT = "Rabbit_Tunnel"   # all non-0 enemies allways flee
    __INF_RESOURCES = "Rich"    # you're always at max resources (keys, energy)
    __MAP_REVEAL = "Illuminati"     # all areas are visible
    __OBSTACLE_IGNORE = "Obstacle-Iqnor"    # obstacles are walkable
    __MESSAGE_IGNORE = "Lequsi"  # can only see messages via history (including options and help menus)
    __NONE = "n0n3"
    __CHEATS = {
        __GOD_MODE: False,
        __INF_EDITS: False,
        __SCARED_RABBIT: False,
        __INF_RESOURCES: False,
        __MAP_REVEAL: False,
        __OBSTACLE_IGNORE: False,
        __MESSAGE_IGNORE: False,
    }

    __allow_cheats = False
    __cheated = False
    __popup = None
    __input_popup = None

    @staticmethod
    def init(popup_callback: Callable[[str, str, int, int], None],
             input_popup_callback: Callable[[str, int, Callable[[str], None]], None], deactivate_cheats: bool = True,
             allow_cheats: bool = False):
        CheatConfig.__allow_cheats = allow_cheats
        CheatConfig.__cheated = False
        CheatConfig.__popup = popup_callback
        CheatConfig.__input_popup = input_popup_callback
        # deactivate cheats if we are not debugging
        if deactivate_cheats:
            for key in CheatConfig.__CHEATS:
                CheatConfig.__CHEATS[key] = False

    @staticmethod
    def did_cheat() -> bool:
        return CheatConfig.__cheated

    @staticmethod
    def in_god_mode() -> bool:
        return CheatConfig.__CHEATS[CheatConfig.__GOD_MODE]

    @staticmethod
    def has_infinite_edits() -> bool:
        return CheatConfig.__CHEATS[CheatConfig.__INF_EDITS]

    @staticmethod
    def is_scared_rabbit() -> bool:
        return CheatConfig.__CHEATS[CheatConfig.__SCARED_RABBIT]

    @staticmethod
    def got_inf_resources() -> bool:
        return CheatConfig.__CHEATS[CheatConfig.__INF_RESOURCES]

    @staticmethod
    def revealed_map() -> bool:
        return CheatConfig.__CHEATS[CheatConfig.__MAP_REVEAL]

    @staticmethod
    def ignore_obstacles() -> bool:
        return CheatConfig.__CHEATS[CheatConfig.__OBSTACLE_IGNORE]

    @staticmethod
    def ignore_dialogue(importance: PopupConfig.Importance) -> bool:
        if importance <= PopupConfig.Importance.Dialogue:
            return CheatConfig.__CHEATS[CheatConfig.__MESSAGE_IGNORE]
        else:
            return False  # messages more important than dialogue are not ignored

    @staticmethod
    def cheat_input():
        if CheatConfig.__allow_cheats and CheatConfig.__input_popup is not None:
            CheatConfig.__input_popup("Input your Cheat:", PyCuiColors.BLACK_ON_RED, CheatConfig.__use_cheat)

    @staticmethod
    def cheat_list():
        text = ""
        for key in CheatConfig.__CHEATS:
            text += f"{key}: \t\t"
            if CheatConfig.__CHEATS[key]:
                text += "Active\n"
            else:
                text += "Inactive\n"
        CheatConfig.__popup("List of Cheats", text, PopupConfig.default_position(), PopupConfig.default_color())

    @staticmethod
    def __use_cheat(code: str) -> bool:
        ret = False
        if code == CheatConfig.__ALL or code == CheatConfig.__NONE:
            for key in CheatConfig.__CHEATS:
                CheatConfig.__CHEATS[key] = code == CheatConfig.__ALL
            ret = True
        elif code in CheatConfig.__CHEATS:
            CheatConfig.__CHEATS[code] = not CheatConfig.__CHEATS[code]
            ret = True
        elif code.lower().strip() == "userdata":
            try:
                open_folder(PathConfig.user_data_path())
            except Exception as ex:
                CheatConfig.__popup("Error", f"Failed to open folder at {PathConfig.user_data_path()}: {ex}",
                                    PopupConfig.default_position(), PopupConfig.default_color())

        if ret:
            CheatConfig.__popup("Cheats", f"Successfully used the Cheat \"{code}\"", PopupConfig.default_position(),
                                PopupConfig.default_color())
            CheatConfig.__cheated = True
        else:
            CheatConfig.__popup("Cheats", "This is not a valid Cheat!", PopupConfig.default_position(),
                                PopupConfig.default_color())
        return ret

    @staticmethod
    def use_cheat(code: str) -> bool:
        assert TestConfig.is_active(), "You are not allowed to cheat outside of testing!"
        return CheatConfig.__use_cheat(code)


class GameplayConfig:
    class _KnowledgeMode(enum.Enum):
        Newbie = 0
        Experienced = 1

    __KNOWLEDGE_MODE: _KnowledgeMode = _KnowledgeMode.Newbie

    @staticmethod
    def set_newbie_mode():
        GameplayConfig.__KNOWLEDGE_MODE = GameplayConfig._KnowledgeMode.Newbie

    @staticmethod
    def set_experienced_mode():
        GameplayConfig.__KNOWLEDGE_MODE = GameplayConfig._KnowledgeMode.Experienced

    @staticmethod
    def set_knowledge_mode(km: int) -> bool:
        if km == 0:
            GameplayConfig.set_newbie_mode()
            return True
        elif km == 1:
            GameplayConfig.set_experienced_mode()
            return True
        return False

    @staticmethod
    def get_knowledge_mode() -> int:
        return GameplayConfig.__KNOWLEDGE_MODE.value

    @staticmethod
    def is_newbie_mode():
        return GameplayConfig.__KNOWLEDGE_MODE == GameplayConfig._KnowledgeMode.Newbie

    @staticmethod
    def is_experienced_mode():
        return GameplayConfig.__KNOWLEDGE_MODE == GameplayConfig._KnowledgeMode.Experienced

    @staticmethod
    def get_option_value(option: Options, convert: Optional[bool] = None, ignore_test_config: Optional[bool] = None):
        return OptionsManager.get_option_value(option, convert, ignore_test_config)


class PuzzleConfig:
    BOSS_FLEE_ENERGY = 10
    BOSS_FAIL_DAMAGE = 5

    @staticmethod
    def calculate_appearance_chance(eid: int) -> float:
        return 0.1 * (10 - eid)  # the lower the id the higher the chance of a puzzle to appear

    @staticmethod
    def calculate_flee_chance(eid: int) -> float:
        return PuzzleConfig.calculate_appearance_chance(eid)

    @staticmethod
    def calculate_flee_energy(eid: int) -> int:
        return max(1, eid * 2)


class ScoreConfig:
    # sometimes we cannot provide an expected number of gates (so value = 0) so we need a default to work with
    __DEFAULT_EXPECTED_GATES: int = 5
    _BONUS_FACTOR: float = 2
    _PENALTY_FACTOR: float = 0.7
    _CHECKS_BONUS_MULT: float = 0.2
    _CHECKS_PENALTY_MULT: float = 1.2
    _USED_WEIGHT: float = 0.6
    _CHECKS_WEIGHT: float = 1 - _USED_WEIGHT
    # base score one gets for solving a puzzle etc.
    _BASE_SCORE: int = 100
    _PUZZLE_MULT: float = 1
    _RIDDLE_MULT: float = 1.5
    _CHALLENGE_MULT: float = 1.5
    # bonus score one can get for solving a puzzle etc. in a low amount of steps
    __PUZZLE_BONUS: int = int(_BASE_SCORE * _PUZZLE_MULT)
    __RIDDLE_BONUS: int = int(_BASE_SCORE * _RIDDLE_MULT)
    __CHALLENGE_BONUS: int = int(_BASE_SCORE * _CHALLENGE_MULT * 1.5)

    @staticmethod
    def _f_bonus(ratio: float, factor: float) -> float:
        # ratio should be < 1
        # gives a higher value the smaller ratio is
        return 1 + (1 + factor) * 2 ** (-ratio) - (1 + factor) * 0.5  # 0.5=2**-1, so to normalize it to 1 for ratio==1

    @staticmethod
    def _f_penalty(ratio: float, factor: float) -> float:
        # ratio should be < 1
        # gives a smaller value the smaller ratio is
        return (1 + factor) ** ratio - factor

    @staticmethod
    def _get_ratio(checks: int, used_gates: int, expected_gates: int) -> float:
        """
        Returns a 0 <= value < 2.5 that is higher the better checks and used_gates performs compared to expected_gates.
        This means the lower checks and used_gates are than expected_gates the higher the returned value. Let us
        distinguish three main cases:
            - checks == used_gates == expected_gates: The returned ratio is 1.0 since the puzzle was solved as expected.
            - used_gates < expected_gates: A better solution was found, so a ratio >1.0 is returned. The penalty for needing many checks is kept low.
            - used_gates > expected_gates: A worse solution was found, so a ratio <1.0 is returned. The penalty for needing many checks is increased.
        Keep in mind that checks can never be smaller than used_gates. Therefore, if no combined gates are used
        expected_gates is also the number of expected checks.

        :param checks: how often it was checked if the target was reached
        :param used_gates: how many gates were used to reach the target
        :param expected_gates: how many gates were expected to be needed to reach the target

        :return: a ratio determining how good a puzzle was solved
        """
        assert not (used_gates is None and expected_gates is not None), f"used_gates is None but expected_gates is " \
                                                                        f"{expected_gates}!"
        if used_gates is None or used_gates <= 0:
            used_gates = 1
        if expected_gates is None or expected_gates <= 0:
            # just skip possible bonus since this should only happen for unimportant puzzles
            # and this way we still get a nice curve only depending on checks
            expected_gates = used_gates
        bonus_factor = ScoreConfig._BONUS_FACTOR
        penalty_factor = ScoreConfig._PENALTY_FACTOR

        # expected_gates is also the expected number of checks since you cannot have less checks than gates
        # todo: above is False if we implement combining gates! But good combinations would be rewarded, so it should
        #  be fine
        if used_gates < expected_gates:
            used_exp_val = ScoreConfig._f_bonus(used_gates / expected_gates, bonus_factor)
            # shrink penalty if we used fewer gates
            penalty_factor *= ScoreConfig._CHECKS_BONUS_MULT
        elif used_gates > expected_gates:
            used_exp_val = ScoreConfig._f_penalty(expected_gates / used_gates, penalty_factor)
            # increase penalty if we used more gates (> 1 could lead to negative numbers though)
            penalty_factor = penalty_factor * ScoreConfig._CHECKS_PENALTY_MULT
        else:
            used_exp_val = 1

        if checks < expected_gates:
            checks_exp_val = ScoreConfig._f_bonus(checks / expected_gates, bonus_factor)
        elif checks > expected_gates:
            checks_exp_val = ScoreConfig._f_penalty(expected_gates / checks, penalty_factor)
        else:
            checks_exp_val = 1

        # in our value range it's not possible to get negative values, but theoretically it is
        return max(used_exp_val * ScoreConfig._USED_WEIGHT + checks_exp_val * ScoreConfig._CHECKS_WEIGHT, 0)

    @staticmethod
    def get_puzzle_score(checks: int, used_gates: int, expected_gates: int) -> int:
        """
        Returns a base score plus bonus score depending on how well the puzzle was solved.

        :param checks: how many steps where takes to reach the target
        :param used_gates: how many gates where used in the final circuit
        :param expected_gates: how many gates where used to create the target StateVector

        :return: a score that is bigger the lower checks and used_gates are compared to expected_gates
        """
        ratio = ScoreConfig._get_ratio(checks, used_gates, expected_gates)
        return ScoreConfig._BASE_SCORE + int(ScoreConfig.__PUZZLE_BONUS * ratio)

    @staticmethod
    def compute_time_bonus(score: int, duration: int) -> int:
        # ratio is exactly 0.5 for a duration of 0s, and then slowly falls off
        # 0 is reached only for an infinite duration
        ratio = 0.5 * 1.2**(-duration / 100)
        return max(int(score * ratio), 0)   # make sure to never return something smaller 0


class QuantumSimulationConfig:
    DECIMALS = 3
    COMPLEX_DECIMALS = 2  # complex numbers need more space and therefore might show fewer decimals
    TOLERANCE = 0.1
    MAX_SPACE_PER_NUMBER = 1 + 1 + 1 + DECIMALS  # sign + "0" + "." + DECIMALS
    MAX_SPACE_PER_COMPLEX_NUMBER = 1 + 1 + COMPLEX_DECIMALS + 1 + 1 + COMPLEX_DECIMALS + 1  # sign, . & j and decimals
    MAX_PERCENTAGE_SPACE = 3  # the maximum (100%) has three digits


class InstructionConfig:
    MAX_ABBREVIATION_LEN = 3
    COMB_GATE_NAME_MIN_CHARACTERS = 1
    COMB_GATE_NAME_MAX_CHARACTERS = 7
    COMB_GATE_MIN_GATE_NUM = 1
    COMB_GATE_MAX_GATE_NUM = 10


class GateType(enum.Enum):
    # unique by their short name
    IGate = "I", "Identity", set(), 0, \
        "An I Gate or Identity Gate doesn't alter the Qubit in any way. It can be used as a placeholder."
    XGate = "X", "Pauli X", {"Pauli-X", "NOT"}, 1, \
        "In the classical world an X Gate corresponds to an inverter or Not Gate.\n" \
        "It swaps the amplitudes of |0> and |1>.\n" \
        "In the quantum world this corresponds to a rotation of 180° along the x-axis, hence the name X Gate."
    SXGate = "SX", "Square Root X", {"Sqrt X"}, 4, \
        "An SX Gate is the square root of an X Gate. This means multiplying two SX Gates (i.e., placing them in " \
        "series) results in an X Gate. Therefore, it corresponds to a rotation of 90° along the x-axis."
    YGate = "Y", "Pauli Y", {"Pauli-Y"}, 4, \
        "A Y Gate rotates the Qubit along the y-axis by 180°."
    ZGate = "Z", "Pauli Z", {"Pauli-Z"}, 4, \
        "A Z Gate rotates the Qubit along the z-axis by 180°."
    HGate = "H", "Hadamard", set(), 2, \
        "The Hadamard Gate is often used to bring Qubits into Superposition."

    SGate = "S", "Phase", {"P", "Phase Shift S"}, 3, \
        "The S Gate can change the phase of a qubit by multiplying its |1> with j (note that this does not alter " \
        "the probability of measuring |0> or |1>!). It is equivalent to a rotation along the z-axis by 90°."
    RYGate = "RY", "Rotational Y", {"Rot Y"}, 4, \
        "The RY Gate conducts a rotation along the y-axis by a certain angle. In our case the angle is 90°."
    RZGate = "RZ", "Rotational Z", {"Rot Z", "Phase Shift Z", "Phase Flip"}, 3, \
        "The RZ Gate conducts a rotation along the z-axis by a certain angle. In our case the angle is 90°."

    SwapGate = "SW", "Swap", set(), 1, \
        "As the name suggests, Swap Gates swap the amplitude between two Qubits."
    CXGate = "CX", "Controlled X", {"CNOT", "Controlled NOT"}, 1, \
        "Applies an X Gate onto its second Qubit (=target) if its first Qubit (=control) is 1."
    CYGate = "CY", "Controlled Y", set(), 4, \
        "Applies a Y Gate onto its second Qubit (=target) if its first Qubit (=control) is 1."
    CZGate = "CZ", "Controlled Z", set(), 4, \
        "Applies a Z Gate onto its second Qubit (=target) if its first Qubit (=control) is 1."
    CHGate = "CH", "Controlled H", {"Controlled Hadamard"}, 3, \
        "Apples an H Gate onto its second Qubit (=target) if its first Qubit (=control) is 1."

    Combined = "Q", "Qombined", set(), 0, \
        "This gate is a combination of multiple gates fused into one."

    Debug = "de", "Debug", set(), 0, "Only use for debugging!"  # used to test spacing

    def __init__(self, short_name: str, full_name: str, other_names: Set[str], difficulty: int, description: str):
        self.__short_name = short_name
        self.__full_name = full_name
        self.__other_names = other_names
        self.__difficulty = difficulty  # how difficult it is to use this gate
        self.__description = description

    @property
    def short_name(self) -> str:
        return self.__short_name

    @property
    def full_name(self) -> str:
        return self.__full_name + " Gate"

    @property
    def has_other_names(self) -> bool:
        return len(self.__other_names) > 0

    @property
    def difficulty(self) -> int:
        return self.__difficulty

    @property
    def description(self) -> str:
        return self.__description

    def is_in_names(self, name: str) -> bool:
        names = {self.__short_name, self.__full_name}
        for other_name in self.__other_names: names.add(other_name)

        if name in names:
            return True
        name = name.lower()
        for n in names:
            if name == n.lower():
                return True
        return False

    def get_other_names(self, separator: str = ", ") -> str:
        return separator.join(self.__other_names)


class ExpeditionConfig:
    DEFAULT_QUEUE_SIZE = 0
    MIN_DIFFICULTY = 1          # minimum difficulty level used for Expeditions
    MIN_NUM_OF_GATES_GOAL = 3   # minimum number of gates we try to pick for non-custom expeditions


class WordingConfig:
    QUANTUM_NEWBIE = "Quantum Newbie"
    QUANTUM_EXPERIENCED = "Quantum Experienced"
